/*
** POSIX library for Lua 4.0. Based on original by Claudio Terra for Lua 3.x
*/

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <time.h>
#include <unistd.h>
#include <utime.h>

#define _POSIX_SOURCE 1

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

#define DEBUG
#undef DEBUG

void lua_poslibopen (lua_State *L);

#define POS_POSLIB_VERSION	"1.4, October, 2000"
#define POS_POSLIB_VERSNUM	1

#include "modemuncher.c"

static FILE *posL_getfile (lua_State *L, int o)
{
	int t;
	if (!lua_isuserdata(L,o)) return NULL;
	lua_getglobal(L, "_STDIN");
	t=lua_tag(L,-1);
	lua_pop(L,1);
	if (t!=lua_tag(L,o))
		return NULL;
	else
		return (FILE*)lua_touserdata(L, o);
}

static void storestring(lua_State *L, const char *name, const char *value)
{
	lua_pushstring(L, name);
	lua_pushstring(L, value);
	lua_settable(L, -3);
}

static void storenumber(lua_State *L, const char *name, double value)
{
	lua_pushstring(L, name);
	lua_pushnumber(L, value);
	lua_settable(L, -3);
}

static void storetable(lua_State *L, const char *index, int subtable)
{
	lua_pushstring(L, index);
	lua_pushvalue(L, subtable);
	lua_settable(L, -3);
}

static int lua_recallobject(lua_State *L, int table, char *index)
{
	lua_pushstring(L, index);
	lua_gettable(L, table);
	return -1;
}

static int pos_makedatetable(lua_State *L, time_t atime)
{
	struct tm* t=localtime(&atime);
	lua_newtable(L);
	storenumber(L, "sec", t->tm_sec);
	storenumber(L, "min", t->tm_min);
	storenumber(L, "hour", t->tm_hour);
	storenumber(L, "day", t->tm_mday);
	storenumber(L, "month", t->tm_mon + 1);
	storenumber(L, "year", t->tm_year + 1900);
	return lua_gettop(L);
}

static int pos_pusherr(lua_State *L) /* taken from Lua IOlib 2.5 */
{
	lua_pushnil(L);
#ifndef NOSTRERROR
	lua_pushstring(L, strerror(errno));
#else
	lua_pushstring(L, "system unable to define the error");
#endif
	return 2;
}

static int pushresult (lua_State *L, int i) /* taken from Lua IOlib 2.5 */
{
  if (i)
  {
    lua_pushuserdata(L, NULL);
    return 1;
  }
  else
  {
   return pos_pusherr(L);
  }
}

static int pos_dir (lua_State *L)
{
	DIR *d = NULL;
	
	switch (lua_type(L, 1))
	{
		case LUA_TNONE:
		d=opendir(".");
		break;
		
		case LUA_TSTRING:
		d=opendir(lua_tostring(L, 1));
		break;
		
		default:
		lua_error(L, "optional argument to function `dir' must be string");
	}
	if (d == NULL)
	{
		return pos_pusherr(L);
	}
	else
	{
		int i = 0;
		struct dirent *entry;
		lua_newtable(L);
		while ((entry = readdir(d)) != NULL)
		{
			lua_pushnumber(L, ++i);
			lua_pushstring(L, entry->d_name);
			lua_settable(L, -3);
		}
		closedir(d);
		lua_pushstring(L, "n");
		lua_pushnumber(L, i);
		lua_settable(L, -3);

		return 1;
	}
}


static int pos_getcwd(lua_State *L)
{
	char buf[500];
	
	if (!getcwd(buf, sizeof(buf)))
	{
		return pos_pusherr(L);
	}
	else
	{
		lua_pushstring(L, buf);
		return 1;
	}
}


static int pos_chdir(lua_State *L)
{
	const char *path = luaL_check_string(L, 1);
	
	if (!chdir(path))
	{
		lua_pushstring(L, path);
		return 1;
	}
	else
	{
		return pos_pusherr(L);
	}
}

static int pos_rmdir (lua_State *L) /* rmdir(path) */
{
	const char *path=luaL_check_string(L, 1);
	if (rmdir(path))
	{
		return pos_pusherr(L);
	}
	else
	{
		lua_pushuserdata(L, NULL);
		return 1;
	}
}

static int pos_remove (lua_State *L) /* remove(path) */
{
	const char *path=luaL_check_string(L, 1);
	if (remove(path))
	{
		return pos_pusherr(L);
	}
	else
	{
		lua_pushuserdata(L, NULL);
		return 1;
	}
}

static int pos_unlink (lua_State *L) /* unlink(path) */
{
	const char *path=luaL_check_string(L, 1);
	if (unlink(path))
	{
		return pos_pusherr(L);
	}
	else
	{
		lua_pushuserdata(L, NULL);
		return 1;
	}
}

static int pos_link (lua_State *L) /* link(existing, new) */
{
	const char *existing=luaL_check_string(L, 1);
	const char *newlink=luaL_check_string(L, 2);
	if (link(existing, newlink))
	{
		return pos_pusherr(L);
	}
	else
	{
		lua_pushuserdata(L, NULL);
		return 1;
	}
}

static int pos_access(lua_State *L)
{
	int amode=0;
	const char *path = luaL_check_string(L, 1);

	if (lua_gettop(L)==1)
	 	amode=F_OK;
	else
	{
		const char *amodestr=luaL_check_string(L, 2);
		for (; *amodestr!=0 ; amodestr++)
			switch (*amodestr)
			{
				case 'r':
				amode |= R_OK;
				break;
				
				case 'w':
				amode |= W_OK;
				break;
				
				case 'x':
				amode |= X_OK;
				break;
				
				case 'f':
				amode |= F_OK;
				break;
			
				/* ignore spaces */
				case ' ':
				break;
				
				default:
				lua_error(L, "bad access mode argument to function `access'");
			}
	}
	if (access(path, amode)==-1) return pos_pusherr(L); /* no access*/
	else lua_pushuserdata(L, NULL); /* access */
	return 1;
}


static int pos_mkdir(lua_State *L)
{
	int i;
	const char *path = luaL_check_string(L, 1);

	mode_t umask_value = umask(0);
	mode_t base_mode = ~umask_value;
	mode_t theMode=base_mode;
	
	switch (lua_type(L, 2))
	{
		case LUA_TSTRING:
		{
			const char *modestr=lua_tostring(L, 2);

			i = mode_munch(&theMode, modestr);
			if (i)
			{
				umask(umask_value);
				lua_error(L, "bad mode change attempt in `mkdir'");
			}
		}
		break;

		case LUA_TNONE:
		break;
		
		
		default:
		{
			umask(umask_value);
			lua_error(L, "optional second argument to function `mkdir' must be string");
		}
	}
	i=mkdir(path, theMode);
	umask(umask_value);	
	if (i)
		return pos_pusherr(L);
	else
	{
		lua_pushnumber(L, theMode);
		return 1;
	}
}

static int pos_chmod(lua_State *L)
{
	mode_t theMode, umask_value, base_mode;
	struct stat result;
	int i;
	const char *modestr;
	const char *path = luaL_check_string(L, 1);
	modestr = luaL_check_string(L, 2);

	umask_value = umask(0);
	if (stat(path, &result)) return pos_pusherr(L); /* stat failed */

	base_mode = result.st_mode;
	
	theMode = base_mode;
	if (mode_munch(&theMode, modestr))
	{
		umask(umask_value);
		lua_error(L, "bad mode change attempt in `chmod'");
	}
	i=chmod(path, theMode);
	umask(umask_value); /* restore old mask */
	if (i)
		return pos_pusherr(L);
	else
	{
		lua_pushnumber(L, theMode);
		return 1;
	}
}

static int pos_chgrp(lua_State *L)
{
	int uid, gid;
	struct stat result;
	const char *path = luaL_check_string(L, 1);
	gid = luaL_check_number(L, 2);

	if (stat(path, &result)) return pos_pusherr(L); /* use stat to get current file uid and gid */
	uid = result.st_uid;

	return pushresult(L, chown(path, uid, gid) == 0);
}

static int pos_chown(lua_State *L)
{
	int uid, gid;
	struct stat result;
	const char *path = luaL_check_string(L, 1);
	uid = luaL_check_number(L, 2);

	if (stat(path, &result)) return pos_pusherr(L); /* use stat to get current file uid and gid */
	gid = result.st_gid;

	return pushresult(L, chown(path, uid, gid) == 0);
}

static int pos_umask(lua_State *L) /* umask([mode]) */
{
	mode_t theMode, umask_value, base_mode;
	char tmp[10];

	umask_value = umask(0);
	theMode = base_mode = ~umask_value;
	
	if ( lua_gettop(L)!=0 )
	{	
		if (lua_isstring(L, 1)) /* user gave mode string to be munched */
			{
				const char *modestr=lua_tostring(L, 1);
				if (mode_munch(&theMode, modestr))
				{
					
					lua_error(L, "bad mode change attempt in `umask'");
				}
			}
			else 
			{
				umask(umask_value);
				lua_error(L, "argument to function `umask' must be string");
			}
			/* user gave bad param type */
	}
	theMode &= ~(S_ISUID | S_ISGID); /* turn off suid and sgid flags */
	umask(~theMode);
#ifdef DEBUG
printf("mask is %u\n", (unsigned int)theMode);
#endif

	modechopper(theMode, tmp);
	lua_pushstring(L, tmp);
	return 1;
}

static time_t pos_maketime(lua_State *L, int table, time_t default_time)
{
	struct tm *tmptr, tmptm;

	tmptr=localtime(&default_time); /* default_time, chopped */
	/* we'll assemble dates into tmptm */
	tmptm.tm_isdst = tmptr->tm_isdst; /* setting up daylight savings time */
	
	/* make access time */
	if (lua_isnil(L,table))
		return default_time;
	else if (lua_istable(L, table)) 
	{
		int o;
		o = lua_recallobject(L, table, "sec");
		if (lua_isnil(L, o)) tmptm.tm_sec = tmptr->tm_sec;
		else tmptm.tm_sec=lua_tonumber(L, o);
		
		o = lua_recallobject(L, table, "min");
		if (lua_isnil(L, o)) tmptm.tm_min = tmptr->tm_min;
		else tmptm.tm_min=lua_tonumber(L, o);

		o = lua_recallobject(L, table, "hour");
		if (lua_isnil(L, o)) tmptm.tm_hour = tmptr->tm_hour;
		else tmptm.tm_hour=lua_tonumber(L, o);

		o = lua_recallobject(L, table, "day");
		if (lua_isnil(L, o)) tmptm.tm_mday = tmptr->tm_mday;
		else tmptm.tm_mday=lua_tonumber(L, o);

		o = lua_recallobject(L, table, "month");
		if (lua_isnil(L, o)) tmptm.tm_mon = tmptr->tm_mon;
		else tmptm.tm_mon=(lua_tonumber(L, o) - 1);

		o = lua_recallobject(L, table, "year");
		if (lua_isnil(L, o)) tmptm.tm_year = tmptr->tm_year;
		else tmptm.tm_year=(lua_tonumber(L, o)-1900);
		return mktime(&tmptm);
	}
	return (time_t)-1; /* parameter must be noobject or table */
}

/*
     tm_year of the tm structure must be for year 1970 or  later.
     Calendar times before 00:00:00 UTC, January 1, 1970 or after
     03:14:07 UTC, January 19, 2038 cannot be represented.
*/


static int pos_fdate (lua_State *L)
{
	struct tm *tm;
	char b[BUFSIZ];
	time_t newtime;
	const char *s = luaL_opt_string(L, 1, "%c");
	
	if ((newtime = pos_maketime(L, 2, time(NULL))) == (time_t)-1)
		lua_error(L, "invalid time parameter to function `date'");

	tm = localtime(&newtime);
	if (strftime(b,sizeof(b),s,tm)) lua_pushstring(L, b);
	else lua_error(L, "invalid `date' format");
	return 1;
}

static int pos_currdate(lua_State *L)
{
	pos_makedatetable(L, time(NULL));
	return 1;
}


static int pos_utime(lua_State *L) /* utime(path, [actime], [modtime]) */
{
	time_t currtime;
	struct utimbuf timbuf;

	const char *path = luaL_check_string(L, 1);
	currtime = time(NULL); /* get current time */
	timbuf.actime = pos_maketime(L, 2, currtime);
	timbuf.modtime = pos_maketime(L, 3, currtime);
	if (timbuf.actime == (time_t)-1 || timbuf.modtime == (time_t)-1)
	{
		lua_pushnil(L);
		lua_pushstring(L, "invalid time parameter to function `utime'");
		return 2;
	}
	return pushresult(L, utime(path, &timbuf) == 0);
}



static const char* procIdVect[] =
{
 	"egid", "euid", "gid", "uid", "pgrp", "pid", "ppid", NULL
};

typedef	uid_t (*functionArray)(void);

static functionArray functionList[] =
{
	(functionArray)getegid,
	(functionArray)geteuid,
	(functionArray)getgid,
	(functionArray)getuid,
	(functionArray)getpgrp,
	(functionArray)getpid,
	(functionArray)getppid,
	NULL
};

static int pos_getprocessid(lua_State *L)
{
	switch (lua_type(L,1))
	{
		case LUA_TSTRING:
		{
			int i;
			i=luaL_findstring(lua_tostring(L, 1), procIdVect);
			if (i==-1) lua_error(L, "unknown selector used in function `getprocessid'");
			else lua_pushnumber(L, functionList[i]());
			/* DEBUG printf("Requested call to %s.\n", procIdVect[i]); */
			/* a valid string selector has been given: we'll return a number */
			return 1;
		}
		break;
	
		case LUA_TNONE:
		{
			/* no argument: we'll return a table with all valid selectors */
			int i=0;
			lua_newtable(L);
			
			while (functionList[i])
			{
				/* DEBUG printf("Requested call to %s.\n", procIdVect[i]); */
				storenumber(L, procIdVect[i], functionList[i]());
				i++;
			}
			return 1;
		}
		
		default:
		lua_error(L, "optional argument to function `getprocessid' must be string");
	}
	return 0;
}


static int pos_getgroup(lua_State *L)
{
	char *e;
	int i=0;
	struct group *result=NULL; /* to avoid warnings */
	
	if ( (lua_gettop(L))==0 )
		lua_error(L, "too few arguments to function `getgroup'");
	
	if (lua_isnumber(L, 1))
		result = getgrgid((gid_t)lua_tonumber(L, 1));
	else if (lua_isstring(L, 1))
		result = getgrnam(lua_tostring(L, 1));
	else lua_error(L, "argument to function `getgroup' must be string or number");
			
	if (result==NULL) return 0; /* CHANGED 961230 */
	
	lua_newtable(L);
	storestring(L, "name", result->gr_name);
	storenumber(L, "gid", result->gr_gid);
	for (i=0; (e=result->gr_mem[i]) != NULL; i++)
	{
		/*printf("%s\n", e);*/
		lua_pushnumber(L, i+1);
		lua_pushstring(L, e);
		lua_settable(L, -3);
	}
	/* store n as total number of items */
	lua_pushstring(L, "n");
	lua_pushnumber(L, i);
	lua_settable(L, -3);
	
	return 1;
}

static int pos_getpasswd(lua_State *L)
{
	struct passwd *result=NULL; /* to avoid warnings */
	
	if ( (lua_gettop(L))==0 )
		lua_error(L, "too few arguments to function `getpasswd'");
	
	if (lua_isnumber(L, 1))
		result = getpwuid((uid_t)lua_tonumber(L, 1));
	else if (lua_isstring(L, 1))
		result = getpwnam(lua_tostring(L, 1));
	else lua_error(L, "argument to function `getpasswd' must be string or number");
			
	if (result==NULL)
	{
		lua_pushnil(L);
		return 1;
	}

	lua_newtable(L);

	storestring(L, "name", result->pw_name);
	storenumber(L, "uid", result->pw_uid);
	storenumber(L, "gid", result->pw_gid);
	storestring(L, "dir", result->pw_dir);
	storestring(L, "shell", result->pw_shell);
	return 1;
}

static int pos_setuid(lua_State *L)
{
	int i=-1;
	
	if ( (lua_gettop(L))==0 )
		lua_error(L, "too few arguments to function `setuid'");
	
	if (lua_isnumber(L, 1))
		i=setuid((uid_t)lua_tonumber(L, 1));
	else if (lua_isstring(L, 1))
	{
		struct passwd *result=getpwnam(lua_tostring(L, 1));
		if (result==NULL)
		{
			return pos_pusherr(L);
		}
		i=setuid(result->pw_uid);
	}
	else lua_error(L, "argument to function `setuid' must be string or number");

	if (i==-1) return pos_pusherr(L);
	else lua_pushuserdata(L, NULL);
	return 1;
}

static int pos_setgid(lua_State *L)
{
	int i=-1;
	
	if ( (lua_gettop(L))==0 )
		lua_error(L, "too few arguments to function `setgid'");
	
	if (lua_isnumber(L, 1))
		i=setgid((gid_t)lua_tonumber(L, 1));
	else if (lua_isstring(L, 1))
	{
		struct group *result=getgrnam(lua_tostring(L, 1));
		if (result==NULL)
		{
			return pos_pusherr(L);
		}
		i=setuid(result->gr_gid);
	}
	else lua_error(L, "argument to function `setgid' must be string or number");

	if (i==-1) return pos_pusherr(L);
	else lua_pushuserdata(L, NULL);
	return 1;
}


static int pos_sleep(lua_State *L)
{
	unsigned int seconds = luaL_check_number(L, 1);

	lua_pushnumber(L, sleep(seconds));
	return 1;
}




static int pos_getenv (lua_State *L)
{
  switch (lua_type(L, 1))
  {
 	case LUA_TNONE:
	{	
		extern char **environ;  /* POSIX variable */
		char *e;
		int i;
		lua_newtable(L);
		for (i=0; (e=environ[i]) != NULL; i++)
		{
			char *eq = strchr(e, '=');
			if (eq == NULL) break;
			*eq = 0;  /* break in two strings */
			lua_pushstring(L, e);		/* name */
			lua_pushstring(L, eq+1);	/* value */
			lua_settable(L, -3);
			*eq = '=';  /* restores original string */
		}
		return 1;
	}
	break;

	case LUA_TSTRING:
	{
		char *c = getenv(lua_tostring(L, 1));
		if (c)
		{
			lua_pushstring(L, c);
			return 1;
		}
		else
		{
			lua_pushnil(L);
			lua_pushstring(L, "environment variable not defined");
			return 2;
		}
	}
	break;

	default:
	lua_error(L, "optional argument to function `getenv' must be string");
	return 0;
   }
}


static int pos_poslibvers(lua_State *L)
{
	lua_pushstring(L, POS_POSLIB_VERSION);
	lua_pushnumber(L, POS_POSLIB_VERSNUM);
	return 2;
}

static int pos_uname(lua_State *L)
{
	luaL_Buffer b;
	struct utsname unamebuf;
	const char *s = luaL_check_string(L, 1);

	if (uname(&unamebuf) == -1) return pos_pusherr(L);

	luaL_buffinit(L, &b);

	for (;*s;s++)
		if (*s != '%')
			luaL_putchar(&b, *s);
		else switch (*++s)
		{
			case '%':
			luaL_putchar(&b, *s);
			break;

			case 's': /* sysname */
			luaL_addstring(&b,unamebuf.sysname);
			break;

			case 'n': /* nodename */
			luaL_addstring(&b,unamebuf.nodename);
			break;

			case 'r': /* release */
			luaL_addstring(&b,unamebuf.release);
			break;

			case 'v': /* version */
			luaL_addstring(&b,unamebuf.version);
			break;

			case 'm': /* machine */
			luaL_addstring(&b,unamebuf.machine);
			break;
			
			default:
			lua_error(L, "invalid format option in function `uname'");
		}
		luaL_pushresult(&b);
		return 1;
}

#ifndef SUNOS4
static const char* sysconfString[] =
{
	"arg_max", "child_max", "clk_tck", "ngroups_max", "stream_max",
	"tzname_max", "open_max", "job_control", "saved_ids","version", NULL
};


static int sysconfConstant[] = 
{
	_SC_ARG_MAX, _SC_CHILD_MAX, _SC_CLK_TCK, _SC_NGROUPS_MAX, _SC_STREAM_MAX,
	_SC_TZNAME_MAX, _SC_OPEN_MAX, _SC_JOB_CONTROL, _SC_SAVED_IDS, _SC_VERSION, -1
};
#else
static const char* sysconfString[] =
{
	"arg_max", "child_max", "clk_tck", "ngroups_max",
	"open_max", "job_control", "saved_ids","version", NULL
};


static int sysconfConstant[] = 
{
	_SC_ARG_MAX, _SC_CHILD_MAX, _SC_CLK_TCK, _SC_NGROUPS_MAX,
	_SC_OPEN_MAX, _SC_JOB_CONTROL, _SC_SAVED_IDS, _SC_VERSION, -1
};
#endif

static int pos_sysconf(lua_State *L)
{
	int name=-1, result;
	
	switch (lua_type(L, 1))
	{
		case LUA_TNIL:
		{
			/* nil parameter: we'll return a table with all valid selectors */
			int i=0;
			lua_newtable(L);
			while (sysconfString[i])
			{
				result = sysconf(sysconfConstant[i]);
				#ifdef DEBUG
				printf("Name=%d, str=%s, r=%d.\n", sysconfConstant[i], sysconfString[i], result);
				#endif
				storenumber(L, sysconfString[i], result);
				i++;
			}
			return 1;
		}
		break;
		
		case LUA_TSTRING:
		name=luaL_findstring(lua_tostring(L, 1), sysconfString);
		if (name==-1) lua_error(L, "unknown selector used in function `sysconf'");
		name=sysconfConstant[name];
		break;
		
		case LUA_TNUMBER: /* user gave numeric constant */
		name=lua_tonumber(L, 1);
		break;
		
		default:	/* user gave bad param type */
		lua_error(L, "argument to function `sysconf' must be string or number");
	}
	/* at this point we have a selector name (int) in `name' */
	result = sysconf(name);
	
	if (result==-1)
	{
		lua_pushnil(L);
		lua_pushstring(L, "selector not defined on the system");
		return 2;
	}
	else
	{
		lua_pushnumber(L, result);
		return 1;
	}
}

static int pos_times (lua_State *L)
{
	struct tms buffer;
	clock_t	elapsed = times(&buffer);

	lua_newtable(L);
	storenumber(L, "elapsed", (double)elapsed/CLK_TCK);
	storenumber(L, "utime", (double)(buffer.tms_utime)/CLK_TCK);
	storenumber(L, "stime", (double)(buffer.tms_stime)/CLK_TCK);
	storenumber(L, "cutime", (double)(buffer.tms_cutime)/CLK_TCK);
	storenumber(L, "cstime", (double)(buffer.tms_cstime)/CLK_TCK);
	return 1;
}

static int pos_clock(lua_State *L)
{
	clock_t val = clock();
	if (val != -1)
	{
		lua_pushnumber(L, (double)val/CLOCKS_PER_SEC);
		return 1;
	}
	else
	{
		lua_error(L, "processor time is unknown");
		return 0;
	}
}

static const char* statString[] = /* Lua selectors for pos_stat */
{
 	"mode", "ino", "dev", "nlink", "uid", "gid", "size", "atime", "mtime", "ctime", NULL
};

static int pos_stat(lua_State *L)
{
	struct stat result;
	int fileref, selector, atime, mtime, ctime;
	char tmp[10];
	int i=0;
	FILE* fl;

	lua_settop(L,2);
	fileref=1;
	selector=2;
	
	switch (lua_type(L, fileref))
	{
		int fildes;
		
		case LUA_TNONE:
		lua_error(L, "too few arguments to function `stat'");
		break;
		
		case LUA_TSTRING:
		i = stat(lua_tostring(L, fileref), &result);
		break;
		
		case LUA_TUSERDATA:
		{
			if (!(fl = posL_getfile(L, fileref)))
			lua_error(L, "userdata first parameter to `stat' must be a file handle");

			fildes = (int)fileno(fl);
			if (fildes == -1)
			{
				return pos_pusherr(L);
			}
			i = fstat(fildes, &result);
		}
		break;
		
		case LUA_TNIL:
		default:
		lua_error(L, "first argument to function `stat' must be string or userdata");

	}
	/* check the error code before returning it */
	/*if (i) pos_error();*/ /* this is to strong; ch. 971203 */
	if (i)
	{
		return pos_pusherr(L);
	}

	/* preparing date/time subtables */
	atime = pos_makedatetable(L, result.st_atime);
	mtime = pos_makedatetable(L, result.st_mtime);
	ctime = pos_makedatetable(L, result.st_ctime);

	/* return mode in rwxrwxrwx format */
	modechopper(result.st_mode, tmp);

	switch (lua_type(L, selector))
	{
		int j;
		
		case LUA_TNIL:
		case LUA_TNONE:
		/* preparing the result table */
		lua_newtable(L);
		storestring(L, statString[0], tmp); /*mode*/
		storenumber(L, statString[1], result.st_ino); /* ino */
		storenumber(L, statString[2], result.st_dev); /* dev */
		storenumber(L, statString[3], result.st_nlink); /* nlink */
		storenumber(L, statString[4], result.st_uid); /* uid */
		storenumber(L, statString[5], result.st_gid); /* gid */
		storenumber(L, statString[6], result.st_size); /* size */
		storetable(L, statString[7], atime); /* atime */
		storetable(L, statString[8], mtime); /* mtime */
		storetable(L, statString[9], ctime); /* ctime */
		return 1;
		break;
		
		case LUA_TSTRING:
		j=luaL_findstring(lua_tostring(L, selector), statString);
		if (j==-1) lua_error(L, "unknown selector used in function `stat'");
		switch (j)
		{
			case 0: lua_pushstring(L, tmp); break;
			case 1: lua_pushnumber(L, result.st_ino); break;
			case 2: lua_pushnumber(L, result.st_dev); break;
			case 3: lua_pushnumber(L, result.st_nlink); break;
			case 4: lua_pushnumber(L, result.st_uid); break;
			case 5: lua_pushnumber(L, result.st_gid); break;
			case 6: lua_pushnumber(L, result.st_size); break;
			case 7: lua_pushvalue(L, atime); break;
			case 8: lua_pushvalue(L, mtime); break;
			case 9: lua_pushvalue(L, ctime); break;
			default:
			lua_error(L, "unknown error ocurred in function `stat'");
		} 
		return 1;
		break;
		
		default:
		lua_error(L, "optional second argument to function `stat' must be string");
	}
	return 0;
}

static int pos_lock (lua_State *L)
{
	struct flock f;
	FILE *fl;
	const char *action;
	int command=F_SETLKW; /* flags*/

	if (!(fl = posL_getfile(L, 1))) lua_error(L, "first parameter to `lock' must be a file handle");
	
  action = luaL_opt_string(L, 2, "w");
  f.l_start = luaL_opt_number(L, 3, 1);
  f.l_len = luaL_opt_number(L, 4, 0);
#if 0
{
	int flags = fcntl(fileno(fl), F_GETFL);
	if (flags == -1) return pos_pusherr(L);
	
	printf("Flags read: %x\n", flags);	
	if (flags & O_RDONLY) printf ("File is read-only.\n");
	if (flags & O_WRONLY) printf ("File is write-only.\n");
	if (flags & O_RDWR) printf ("File is read-write.\n");
	if (flags & O_APPEND) printf ("Append mode.\n");
	printf("Action is '%s'.\n", action);
}
#endif


  switch (*action)
  {
    case 'r' : f.l_type = F_RDLCK; break;
    case 'w' : f.l_type = F_WRLCK; break;
    case 'u' : f.l_type = F_UNLCK; break;
    default: lua_error(L, "invalid mode in function `lock'");
  }
  
  switch (action[1]) /* see if this is a question mark action */
  {
  	case 0: break;
  	case '?': command = F_GETLK; break;
  	default: lua_error(L, "invalid mode in function `lock'");
  }
  
  f.l_pid = 0;
  f.l_whence = SEEK_SET;

/*printf("Achtung! Will call.\n");*/
  
  if (fcntl(fileno(fl), command, &f) != -1)
  {
	  if (command==F_GETLK)
	  {
	    if (f.l_type!=F_UNLCK) lua_pushnumber(L, f.l_pid);
	    else lua_pushnil(L);
	    return 1;
	  }
  }
  return pos_pusherr(L);
}


static const char* confString[] = /* Lua names */
{
 	"link_max", "max_canon", "max_input", "name_max",
 	"path_max", "pipe_buf", "chown_restricted", "no_trunc", "vdisable", NULL
};

static int confConstant[] = /* POSIX constants */
{
	_PC_LINK_MAX, _PC_MAX_CANON, _PC_MAX_INPUT, _PC_NAME_MAX,
	_PC_PATH_MAX, _PC_PIPE_BUF, _PC_CHOWN_RESTRICTED, _PC_NO_TRUNC, _PC_VDISABLE, -1
};

static int pos_pathconf(lua_State *L) /* pathconf(string|userdata, [string|number]) */
{
	int fileref, selector;
	int name=-1;
	int fildes=-1;
	FILE *fl;
	long result;
	
	fileref=1;
	switch (lua_type(L, fileref))
	{
		case LUA_TNONE:
		lua_error(L, "too few arguments to function `pathconf'");
		break;
		
		case LUA_TUSERDATA:

		if (!(fl = posL_getfile(L, fileref)))
		lua_error(L, "userdata first parameter to `pathconf' must be a file handle");

		fildes = (int)fileno(fl);
		if (fildes == -1) return pos_pusherr(L);
		break;

		case LUA_TSTRING:
		break;
		
		case LUA_TNIL:
		default:
		lua_error(L, "first argument to function `pathconf' must be string or userdata");
	}
	/* After this switch, we know if fileref is string or userdata by checking */
	/* `fildes'. If it is still -1 we'll use the string, else we'll use fildes itself.*/

	selector=2;
	
	switch (lua_type(L, selector))
	{
		case LUA_TNUMBER: /* user gave numeric constant */
		name=lua_tonumber(L, selector);
		break;
		
		case LUA_TSTRING: /* user gave string constant */
		name=luaL_findstring(lua_tostring(L, selector), confString);
		if (name==-1) lua_error(L, "unknown selector used in function `pathconf'");
		name=confConstant[name];
		break;
		
		case LUA_TNONE:
		{
			/* missing parameter: we'll return a table with all valid selectors */
			int i=0;
			lua_newtable(L); /* opens table */
			
			if (fildes==-1)
			while (confString[i])
			{
				result = pathconf(lua_tostring(L, fileref), confConstant[i]);
				#ifdef DEBUG
				printf("String; name=%d, str=%s, r=%d.\n", confConstant[i], confString[i], result);
				#endif
				/* changed because of IRIX 5.3 */
				if (result!=-1) storenumber(L, confString[i], result);
				i++;
			}
			else while (confString[i])
			{
				result = fpathconf(fildes, confConstant[i]);
				#ifdef DEBUG
				printf("Userdata; name=%d, str=%s, r=%d.\n", confConstant[i], confString[i], result);
				#endif
				/* changed because of IRIX 5.3 */
				if (result!=-1) storenumber(L, confString[i], result);
				i++;
			}
			return 1;
		}
		default: /* user gave bad param type */
		lua_error(L, "second argument to function `pathconf' should be string or number");
		break;
	}
	/* at this point we have a selector name (int) in `name' */
	#ifdef DEBUG
	printf("Numeric name is %d.\n", name);
	#endif
	if (fildes==-1) result = pathconf(lua_tostring(L, fileref), name);
	else result = fpathconf(fildes, name);
	/*if ((result==-1) && errno) pos_error();*/ /* changed because of IRIX 5.3 */
	if ((result==-1) /*&& !errno*/)
	{
		lua_pushnil(L);
		lua_pushstring(L, "error getting limit for the given pathconf selector");
		return 2;
	}
	lua_pushnumber(L, result);
	return 1;
}


static struct luaL_reg poslib[] =
{
	/* Lua Name		Function pointer */

	{"access",		pos_access},
	{"chdir",		pos_chdir},
	{"chgrp",		pos_chgrp},
	{"chmod",		pos_chmod},
	{"chown",		pos_chown},
	{"clock",		pos_clock},
	{"currdate",		pos_currdate},
	{"date",		pos_fdate},
	{"dir",			pos_dir},
	{"getcwd",		pos_getcwd},
	{"getenv",		pos_getenv},
	{"getgroup",		pos_getgroup},
	{"getpasswd",		pos_getpasswd},
	{"getprocessid",	pos_getprocessid},
	{"link",		pos_link},
	{"lock",		pos_lock},
	{"mkdir",		pos_mkdir},
	{"pathconf",		pos_pathconf},
	{"poslibvers",		pos_poslibvers},
	{"remove",		pos_remove},
	{"rmdir",		pos_rmdir},
	{"setgid",		pos_setgid},
	{"setuid",		pos_setuid},
	{"sleep",		pos_sleep},
	{"stat",		pos_stat},
	{"sysconf",		pos_sysconf},
	{"times",		pos_times},
	{"umask",		pos_umask},
	{"uname",		pos_uname},
	{"unlink",		pos_unlink},
	{"utime",		pos_utime},

};

void lua_poslibopen (lua_State *L)
{
	luaL_openl(L, poslib);
}
